@@ -1,10 +1,12 @@
-#!/usr/bin/python
+#!/usr/bin/python3
 # -*- coding: utf-8 -*-
 # vim:et sts=4 sw=4
 #
 # ibus-table - The Tables engine for IBus
 #
 # Copyright (c) 2008-2009 Yu Yuwei <acevery@gmail.com>
+# Copyright (c) 2009-2014 Caius "kaio" CHANCE <me@kaio.net>
+# Copyright (c) 2012-2014 Mike FABIAN <mfabian@redhat.com>
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -20,8 +22,6 @@
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 #
-# $Id: $
-#
 
 import os
 import sys
@@ -62,51 +62,60 @@ class InvalidTableName(Exception):
         return 'Value of NAME attribute (%s) cannot contain any of %r and must be all ascii' % (self.table_name, _invalid_keyname_chars)
 
 # we use OptionParser to parse the cmd arguments :)
-opt_parser = OptionParser()
+usage = "usage: %prog [options]"
+opt_parser = OptionParser(usage=usage)
 
 opt_parser.add_option( '-n', '--name',
-        action = 'store', dest='name',default = None,
-        help = 'set the database name we will use, default is %default')
+        action = 'store', dest='name',default = '',
+        help = "specifies the file name for the binary database for the IME. The default is '%default'. If the file name of the database is not specified, the file name of the source file before the first '.' will be appended with '.db' and that will be used as the file name of the database.")
 
 opt_parser.add_option( '-s', '--source',
-        action = 'store', dest='source', default = 'xingma.txt.bz2',
-        help = 'tell me which file is the source file of IME, default is %default')
+        action = 'store', dest='source', default = '',
+        help = "specifies the file which contains the source of the IME.  The default is '%default'.")
 
 opt_parser.add_option( '-e', '--extra',
         action = 'store', dest='extra', default = '',
-        help = 'tell me which file is the extra words file for IME, default is %default')
+        help = "specifies the file name for the extra words for the IME.  The default is '%default'.")
 
 opt_parser.add_option( '-p', '--pinyin',
         action = 'store', dest='pinyin', default = '/usr/share/ibus-table/data/pinyin_table.txt.bz2',
-        help = 'tell me which file is the source file of pinyin, default is %default')
+        help = "specifies the source file for the  pinyin.  The default is '%default'.")
 
 opt_parser.add_option( '-o', '--no-create-index',
         action = 'store_false', dest='index', default = True,
-        help = 'do not create index on database, only for distrubution purpose, normal user should not invoke this flag!')
+        help = 'Do not create an index for a database (Only for distrubution purposes, a normal user should not use this flag!)')
 
 opt_parser.add_option( '-i', '--create-index-only',
         action = 'store_true', dest='only_index', default = False,
-        help = 'only create index on exist database')
+        help = 'Only create an index for an existing database. Specifying the file name of the binary database with the -n or --name option is required when this option is used.')
 
 opt_parser.add_option( '-d', '--debug',
         action = 'store_true', dest='debug', default = False,
-        help = 'print extra debug messages')
-
-
+        help = 'Print extra debug messages.')
 
 opts,args = opt_parser.parse_args()
-if not opts.name and opts.only_index:
-    print 'Please give me the database you want to create index on'
-    sys.exit(2)
+if opts.only_index:
+    if not opts.name:
+        opt_parser.print_help()
+        print('\nPlease specify the file name of the database you want to create an index on!')
+        sys.exit(2)
+    if not os.path.exists(opts.name) or not os.path.isfile(opts.name):
+        opt_parser.print_help()
+        print("\nThe database file '%s' does not exist." %opts.name)
+        sys.exit(2)
 
-if not opts.name:
+if not opts.name and opts.source:
     opts.name = os.path.basename(opts.source).split('.')[0] + '.db'
 
+if not opts.name:
+    opt_parser.print_help()
+    print('\nYou need to specify the file which contains the source of the IME!')
+    sys.exit(2)
 
 def main ():
     def debug_print ( message ):
         if opts.debug:
-            print message
+            print(message)
 
     if not opts.only_index:
         try:
@@ -115,8 +124,9 @@ def main ():
             pass
 
     debug_print ("Processing Database")
-    db = tabsqlitedb.tabsqlitedb ( filename = opts.name)
-    #db.db.execute( 'PRAGMA synchronous = FULL; ' )
+    db = tabsqlitedb.tabsqlitedb(filename = opts.name,
+                                 user_db = None,
+                                 create_database = True)
 
     def parse_source (f):
         _attri = []
@@ -124,10 +134,9 @@ def main ():
         _gouci = []
         patt_com = re.compile(r'^###.*')
         patt_blank = re.compile(r'^[ \t]*$')
-        patt_conf = re.compile(r'.*=.*')
-        patt_table = re.compile(r' *([^\s]+) *\t *([^\s]+)\t *[^\s]+ *')
-        patt_gouci = re.compile(r' *[^\s]+ *\t *[^\s]+ *')
-        patt_s = re.compile(r' *([^\s]+) *\t *([\x00-\xff]{3}) *\t *[^\s]+ *')
+        patt_conf = re.compile(r'[^\t]*=[^\t]*')
+        patt_table = re.compile(r'([^\t]+)\t([^\t]+)\t([^t]+)(\t.*)?$')
+        patt_gouci = re.compile(r' *[^\s]+ *\t *[^\s]+ *$')
 
         for l in f:
             if ( not patt_com.match(l) ) and ( not patt_blank.match(l) ):
@@ -135,13 +144,32 @@ def main ():
                     if _patt.match(l):
                         _list.append(l)
                         break
+
         if not _gouci:
-            #user didn't provide goucima, so we use the longest single character encode as the goucima.
+            # The user didn’t provide goucima (goucima = 構詞碼 =
+            # “word formation keys”) in the table source, so we use
+            # the longest encoding for a single character as the
+            # goucima for that character.
+            #
+            # Example:
+            #
+            # wubi-jidian86.txt contains:
+            #
+            #     a         工      99454797
+            #     aaa	工      551000000
+            #     aaaa      工      551000000
+            #     aaad      工期    5350000
+            #     ... and more matches for compounds containing 工
+            #
+            # The longest key sequence to type 工 as a single
+            # character is “aaaa”.  Therefore, the goucima of 工 is
+            # “aaaa” (There is one other character with the same goucima
+            # in  wubi-jidian86.txt, 㠭 also has the goucima “aaaa”).
             gouci_dict = {}
             for line in _table:
-                res = patt_s.match(line)
-                if res:
-                    if gouci_dict.has_key(res.group(2)):
+                res = patt_table.match(line)
+                if res and len(res.group(2)) == 1:
+                    if res.group(2) in gouci_dict:
                         if len(res.group(1)) > len(gouci_dict[res.group(2)]):
                             gouci_dict[res.group(2)] = res.group(1)
                     else:
@@ -160,6 +188,8 @@ def main ():
         patt_yin = re.compile(r'[a-z]+[1-5]')
 
         for l in f:
+            if type(l) != type(u''):
+                l = l.decode('utf-8')
             if ( not patt_com.match(l) ) and ( not patt_blank.match(l) ):
                 res = patt_py.match(l)
                 if res:
@@ -174,7 +204,6 @@ def main ():
         patt_com = re.compile(r'^###.*')
         patt_blank = re.compile(r'^[ \t]*$')
         patt_extra = re.compile(r'(.*)\t(.*)')
-        patt_s = re.compile(r'(.*)\t([\x00-\xff]{3})\t.*')
 
         for l in f:
             if ( not patt_com.match(l) ) and ( not patt_blank.match(l) ):
@@ -185,13 +214,17 @@ def main ():
 
     def pinyin_parser (f):
         for py in f:
-            _zi, _pinyin, _freq = unicode (py,'utf-8').strip ().split()
+            if type(py) != type(u''):
+                py = py.decode('utf-8')
+            _zi, _pinyin, _freq = py.strip().split()
             yield (_pinyin, _zi, _freq)
 
     def phrase_parser (f):
         list=[]
         for l in f:
-            xingma, phrase, freq = unicode (l, "utf-8").strip ().split ('\t')[:3]
+            if type(l) != type(u''):
+                l = l.decode('utf-8')
+            xingma, phrase, freq = l.strip().split('\t')[:3]
             if phrase == 'NOSYMBOL':
                 phrase = u''
             list.append ( (xingma, phrase, int(freq), 0) )
@@ -199,34 +232,62 @@ def main ():
 
     def goucima_parser (f):
         for l in f:
-            zi,gcm = unicode (l, "utf-8").strip ().split ()
+            if type(l) != type(u''):
+                l = l.decode('utf-8')
+            zi,gcm = l.strip().split()
             yield (zi, gcm)
 
     def attribute_parser (f):
         for l in f:
+            if type(l) != type(u''):
+                l = l.decode('utf-8')
             try:
-                attr,val = unicode (l,"utf-8").strip().split ('=')
+                attr,val = l.strip().split('=')
             except:
-                attr,val = unicode (l,"utf-8").strip().split ('==')
-            attr = attr.strip()
-            origin_attr = attr
-            attr = attr.lower()
+                attr,val = l.strip().split('==')
+            attr = attr.strip().lower()
             val = val.strip()
-            if attr == 'name' and not gconf_valid_keyname(val):
-                raise InvalidTableName(val)
             yield (attr,val)
 
     def extra_parser (f):
         list = []
         for l in f:
-            phrase, freq = unicode (l, "utf-8").strip ().split ()
-            try:
-                _tabkey = db.parse_phrase_to_tabkeys(phrase)
-                list.append( (_tabkey,phrase,freq,0) )
-            except:
-                print '\"%s\" would not been added' % phrase.encode('utf-8')
+            if type(l) != type(u''):
+                l = l.decode('utf-8')
+            phrase, freq = l.strip().split ()
+            _tabkey = db.parse_phrase(phrase)
+            if _tabkey:
+                list.append((_tabkey,phrase,freq,0))
+            else:
+                print('No tabkeys found for “%s”, not adding.\n' %phrase)
         return list
 
+    def get_char_prompts(f):
+        '''
+        Returns something like
+
+        ("char_prompts", "{'a': '日', 'b': '日', 'c': '金', ...}")
+
+        i.e. the attribute name "char_prompts" and as its value
+        the string representation of a Python dictionary.
+        '''
+        char_prompts = {}
+        start = False
+        for l in f:
+            if type(l) != type(u''):
+                l = l.decode('utf-8')
+            if re.match(r'^BEGIN_CHAR_PROMPTS_DEFINITION', l):
+                start = True
+                continue
+            if not start:
+                continue
+            if re.match(r'^END_CHAR_PROMPTS_DEFINITION', l):
+                break
+            match = re.search(r'^(?P<char>[^\s]+)[\s]+(?P<prompt>[^\s]+)', l)
+            if match:
+                char_prompts[match.group('char')] = match.group('prompt')
+        return ("char_prompts", repr(char_prompts))
+
     if opts.only_index:
         debug_print ('Only create Indexes')
         debug_print ( "Optimizing database " )
@@ -242,34 +303,37 @@ def main ():
     patt_s = re.compile( r'.*\.bz2' )
     _bz2s = patt_s.match(opts.source)
     if _bz2s:
-        source = bz2.BZ2File ( opts.source, "r" )
+        source = bz2.BZ2File(opts.source, "r").read()
     else:
-        source = file ( opts.source, 'r' )
+        source = open(opts.source, mode='r', encoding='UTF-8').read()
+    source = source.replace('\r\n', '\n')
+    source = source.split('\n')
     # first get config line and table line and goucima line respectively
     debug_print ('\tParsing table source file ')
     attri,table,gouci =  parse_source ( source )
 
-    debug_print ('\t  get attribute of IME :)')
-    attributes = attribute_parser ( attri )
-    debug_print ('\t  add attributes into DB ')
-    db.update_ime ( attributes )
-    db.create_tables ('main')
+    debug_print('\t  get attribute of IME :)')
+    attributes = list(attribute_parser(attri))
+    attributes.append(get_char_prompts(source))
+    debug_print('\t  add attributes into DB ')
+    db.update_ime( attributes )
+    db.create_tables('main')
 
     # second, we use generators for database generating:
     debug_print ('\t  get phrases of IME :)')
     phrases = phrase_parser ( table)
 
     # now we add things into db
-    debug_print ('\t  add phrases into DB ')
-    db.add_phrases ( phrases )
+    debug_print('\t  add phrases into DB ')
+    db.add_phrases(phrases)
 
-    if db.get_ime_property ('user_can_define_phrase').lower() == u'true':
+    if db.ime_properties.get('user_can_define_phrase').lower() == u'true':
         debug_print ('\t  get goucima of IME :)')
         goucima = goucima_parser (gouci)
         debug_print ('\t  add goucima into DB ')
         db.add_goucima ( goucima )
 
-    if db.get_ime_property ('pinyin_mode').lower() == u'true':
+    if db.ime_properties.get('pinyin_mode').lower() == u'true':
         debug_print ('\tLoad pinyin source \"%s\"' % opts.pinyin)
         _bz2p = patt_s.match(opts.pinyin)
         if _bz2p:
@@ -286,7 +350,7 @@ def main ():
     debug_print ("Optimizing database ")
     db.optimize_database ()
 
-    if db.get_ime_property ('user_can_define_phrase').lower() == u'true' and opts.extra:
+    if db.ime_properties.get('user_can_define_phrase').lower() == u'true' and opts.extra:
         debug_print( '\tPreparing for adding extra words' )
         db.create_indexes ('main')
         debug_print ('\tLoad extra words source \"%s\"' % opts.extra)
@@ -298,22 +362,24 @@ def main ():
         debug_print ('\tParsing extra words source file ')
         extraline = parse_extra (extra_s)
         debug_print ('\tPreparing extra words lines')
-        db.cache_goucima()
-        debug_print ('\t  Goucima has been cache to memory')
         extrawds = extra_parser (extraline)
         debug_print( '\t  we have %d extra phrases from source' % len(extrawds))
         # first get the entry of original phrases from
         # phrases-[(xingma, phrase, int(freq), 0)]
         orig_phrases = {}
-        map (lambda x: orig_phrases.update({"%s\t%s"%(x[0],x[1]):x}), phrases )
+        for x in phrases:
+            orig_phrases.update({"%s\t%s"%(x[0],x[1]):x})
         debug_print( '\t  the len of orig_phrases is: %d' % len(orig_phrases) )
         extra_phrases = {}
-        map (lambda x: extra_phrases.update({"%s\t%s" %(x[0],x[1]):x}), extrawds )
+        for x in extrawds:
+            extra_phrases.update({"%s\t%s" %(x[0],x[1]):x})
         debug_print ( '\t  the len of extra_phrases is: %d' % len(extra_phrases) )
         # pop duplicated keys
-        map (lambda x: extra_phrases.pop(x) if orig_phrases.has_key(x) else 0, extra_phrases.keys() )
+        for x in extra_phrases:
+            if x in orig_phrases:
+                extra_phrases.pop(x)
         debug_print( '\t  %d extra phrases will be added' % len(extra_phrases))
-        new_phrases = extra_phrases.values()
+        new_phrases = list(extra_phrases.values())
         debug_print ('\tAdding extra words into DB ')
         db.add_phrases (new_phrases)
         debug_print ("Optimizing database ")
